/*
 * CPU-specific routines originally from U-Boot.
 * See: uboot-xburst/files/arch/mips/cpu/xburst/cpu.c
 * See: u-boot/arch/mips/include/asm/cacheops.h
 *
 * Copyright (C) 2000-2009 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "cpu.h"
#include "sdram.h"

void flush_icache_all(void)
{
	u32 addr, t = 0;

	asm volatile ("mtc0 $0, $28"); /* Clear Taglo */
	asm volatile ("mtc0 $0, $29"); /* Clear TagHi */

	for (addr = KSEG0; addr < KSEG0 + CONFIG_SYS_ICACHE_SIZE;
	     addr += CONFIG_SYS_CACHELINE_SIZE) {
		asm volatile (
			".set mips3\n\t"
			" cache %0, 0(%1)\n\t"
			".set mips2\n\t"
			:
			: "I" (Index_Store_Tag_I), "r"(addr));
	}

	/* invalicate btb */
	asm volatile (
		".set mips32\n\t"
		"mfc0 %0, $16, 7\n\t"
		"nop\n\t"
		"ori %0,2\n\t"
		"mtc0 %0, $16, 7\n\t"
		".set mips2\n\t"
		:
		: "r" (t));
}

void flush_dcache_all(void)
{
	u32 addr;

	for (addr = KSEG0; addr < KSEG0 + CONFIG_SYS_DCACHE_SIZE; 
	     addr += CONFIG_SYS_CACHELINE_SIZE) {
		asm volatile (
			".set mips3\n\t"
			" cache %0, 0(%1)\n\t"
			".set mips2\n\t"
			:
			: "I" (Index_Writeback_Inv_D), "r"(addr));
	}

	asm volatile ("sync");
}

void flush_cache_all(void)
{
	flush_dcache_all();
	flush_icache_all();
}

void handle_error_level(void)
{
	asm volatile(
		"mfc0 $t3, $12\n" /* CP0_STATUS */
		"li $t4, 0xfffffffb\n" /* ERL = 0 */
		"and $t3, $t3, $t4\n"
		"mtc0 $t3, $12\n"
		"nop\n");
}

void enable_interrupts(void)
{
	asm volatile(
		"mfc0 $t3, $12\n" /* CP0_STATUS */
		"li $t4, 0x0000fc01\n" /* IE = enable interrupts */
		"or $t3, $t3, $t4\n"
		"mtc0 $t3, $12\n"
		"nop\n");
}

void init_interrupts(void)
{
	/* Set exception registers. */

	asm volatile(
		"mtc0 $zero, $18\n"	/* CP0_WATCHLO */
		"li $t3, 0x00800000\n"	/* IV = 1 (use 0x80000200 for interrupts) */
		"mtc0 $t3, $13\n"	/* CP0_CAUSE */
		"mfc0 $t4, $12\n" 	/* CP0_STATUS */
		"li $t3, 0xffbfffff\n"	/* BEV=0 */
		"and $t3, $t3, $t4\n"
		"mtc0 $t3, $12\n"
		"nop\n");
}

void init_tlb(void)
{
	asm volatile(
		"li $t0, 0x01ffe000\n"	/* 16MB */
		"mtc0 $t0, $5\n"	/* CP0_PAGEMASK */
		"li $t1, 2\n"		/* index of first randomly-replaced entry */
		"mtc0 $t1, $6\n"	/* CP0_WIRED */

		/* 0x80000000..0x82000000 -> 0x00000000..0x02000000 */

		"mtc0 $zero, $0\n"	/* CP0_INDEX */

		/* Set physical address. */

		"li $t0, 0x0000001f\n"	/* 0x00000000, C=3, dirty, global, valid */
		"mtc0 $t0, $2\n"	/* CP0_ENTRYLO0 */
		"li $t1, 0x0004001f\n"	/* 0x01000000, C=3, dirty, global, valid */
		"mtc0 $t1, $3\n"	/* CP0_ENTRYLO1 */

		/* Set virtual address. */

		"li $t0, 0x80000000\n"	/* 0x80000000, ASID=0 */
		"mtc0 $t0, $10\n"	/* CP0_ENTRYHI */
		"nop\n"

		"tlbwi\n"
		"nop\n");

	map_page(0x00000000, 0x00000000, 16 * 1024 * 1024);
}

void map_page(u32 virtual, u32 physical, u32 pagesize)
{
	u32 start = virtual & 0xffffe000;
	u32 flags = 0x1f; /* C=3, dirty, global, valid */
	u32 lower = ((physical & 0xffffc000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xffffc000) >> 6) | flags;
	u32 pagemask = ((pagesize - 1) & 0xfffff000) << 1;

	asm volatile(
		"mtc0 %3, $5\n"		/* CP0_PAGEMASK */
		"li $t1, 2\n"		/* index of first randomly-replaced entry */
		"mtc0 $t1, $6\n"	/* CP0_WIRED */

		/* Set physical address. */

		"mtc0 %0, $2\n"		/* CP0_ENTRYLO0 */
		"mtc0 %1, $3\n"		/* CP0_ENTRYLO1 */

		/* Set virtual address. */

		"mtc0 %2, $10\n"	/* CP0_ENTRYHI */
		"nop\n"

		"tlbwr\n"
		"nop"
		:
		: "r" (lower), "r" (upper), "r" (start), "r" (pagemask)
		);
}
